use crate::config::WebConfig;
use crate::web::IndexAccess;
use crate::webutil::{internal_error, TemplatedHtml};
use askama::Template;
use axum::extract::{Extension, Query};
use axum::response::IntoResponse;
use serde::Deserialize;

#[derive(Clone, Template)]
#[template(path = "search.html.askama")]
pub struct SearchTemplate {
    pub search_term: String,
    pub results: Vec<SearchResult>,
    pub contact: Vec<(String, String)>,
}

#[derive(Clone)]
pub struct SearchResult {
    pub favicon_url: String,
    pub url: String,
    pub title: String,
    pub excerpt: String,
    pub tags: Vec<String>,
}

#[derive(Clone)]
pub struct ExcerptFragment {
    pub text: String,
    pub mark: bool,
}

pub async fn search_root(Extension(web_config): Extension<WebConfig>) -> impl IntoResponse {
    let seed_config = &web_config.seed_collection;
    TemplatedHtml(SearchTemplate {
        search_term: String::with_capacity(0),
        results: vec![],
        contact: seed_config.contact.clone(),
    })
}

#[derive(Deserialize, Debug, Clone)]
pub struct QueryParameters {
    /// The search query
    q: String,
}

pub async fn search_search(
    web_config: Extension<WebConfig>,
    index_access: Extension<IndexAccess>,
    params: Query<QueryParameters>,
) -> impl IntoResponse {
    search_search_inner(web_config, index_access, params)
        .await
        .map_err(internal_error)
}

pub async fn search_search_inner(
    Extension(web_config): Extension<WebConfig>,
    Extension(index_access): Extension<IndexAccess>,
    Query(params): Query<QueryParameters>,
) -> anyhow::Result<impl IntoResponse> {
    let seed_config = &web_config.seed_collection;

    let raw_results = index_access.backend.query(params.q.clone())?;

    let mut results = Vec::with_capacity(raw_results.len());

    for search_doc in raw_results {
        results.push(SearchResult {
            favicon_url: format!("{:16x}", search_doc.favicon_url_hash),
            url: search_doc.url,
            title: search_doc.title,
            excerpt: search_doc.excerpt,
            tags: search_doc.tags,
        });
    }

    Ok(TemplatedHtml(SearchTemplate {
        search_term: params.q.clone(),
        results,
        contact: seed_config.contact.clone(),
    }))
}
