use anyhow::{bail, Context};
use clap::Parser;
use colour::{blue, yellow_ln};
use env_logger::Env;
use std::collections::HashMap;
use std::fs::File;
use std::io::{BufRead, BufReader};

use patricia_tree::PatriciaMap;
use quickpeep_densedoc::DenseTree;
use quickpeep_index::auxiliary::icon_store::IconStore;
use quickpeep_index::backend::{Backend, BackendIndependentDocument};
use quickpeep_indexer::config::IndexerConfig;
use quickpeep_seed_parser::loader::{
    find_seed_files, seed_loader, Seed, UrlOrUrlPattern, SEED_EXTENSION,
};
use quickpeep_structs::rake_entries::{
    AnalysisAntifeatures, IconEntry, PackRecord, RakedPageEntry, SCHEMA_RAKED_ICONS,
    SCHEMA_RAKED_PAGES,
};
use quickpeep_utils::urls::get_reduced_domain;
use smartstring::alias::CompactString;
use std::path::PathBuf;
use tokio::sync::mpsc::Receiver;
use url::Url;

/// Seeds a raker's queue with URLs
#[derive(Clone, Debug, Parser)]
pub struct Opts {
    #[clap(long = "config")]
    config: Option<PathBuf>,

    rakepacks: Vec<PathBuf>,
}

#[tokio::main]
pub async fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info,qp_indexer=debug")).init();

    let opts: Opts = Opts::parse();

    let config_path = opts
        .config
        .unwrap_or_else(|| PathBuf::from("qp_indexer.toml"));
    let config = IndexerConfig::load(&config_path).context("Failed to load config")?;

    let icon_store = IconStore::open(config.icon_store.as_path())?;

    let seed_files = find_seed_files(config.seed_dir.clone(), SEED_EXTENSION).await?;
    let (seed_tx, seed_rx) = tokio::sync::mpsc::channel(64);
    let handle = tokio::spawn(async move {
        seed_loader(seed_files, &seed_tx).await?;
        Ok(()) as anyhow::Result<()>
    });

    let seed_lookup = build_seed_lookup_table(seed_rx).await?;

    handle.await??;

    let mut indexer_backend = config.open_indexer_backend()?;

    for pack in opts.rakepacks {
        blue!("Indexing: ");
        yellow_ln!("{:?}", pack);

        let file = File::open(&pack)?;
        let decompressor = zstd::stream::Decoder::new(file)?;
        // TODO the decompressor has a buffer already, but we need this to see the end
        let mut buf_reader = BufReader::new(decompressor);
        let schema: String = serde_bare::from_reader(&mut buf_reader)?;

        match schema.as_ref() {
            SCHEMA_RAKED_PAGES => {
                // TODO(unstable): this condition is `.has_data_left()` but it's unstable.
                while buf_reader.fill_buf().map(|b| !b.is_empty())? {
                    handle_page_pack(&mut buf_reader, &seed_lookup, &mut indexer_backend)?;
                }
            }
            SCHEMA_RAKED_ICONS => {
                // TODO(unstable): this condition is `.has_data_left()` but it's unstable.
                while buf_reader.fill_buf().map(|b| !b.is_empty())? {
                    handle_icon_pack(&mut buf_reader, &icon_store)?;
                }
            }
            _ => {
                bail!(
                    "Wrong schema version: wanted e.g. {:?}, got {:?}",
                    SCHEMA_RAKED_PAGES,
                    &schema
                );
            }
        }
    }
    indexer_backend.flush()?;

    Ok(())
}

pub fn handle_page_pack(
    buf_reader: &mut impl BufRead,
    seed_lookup: &SeedLookupTable,
    indexer_backend: &mut Box<dyn Backend>,
) -> anyhow::Result<()> {
    let page_record: PackRecord<RakedPageEntry> = serde_bare::from_reader(buf_reader)?;

    let document = page_record.record.document;

    let article_body = DenseTree::generate_textual_format(&document.body_content, false);
    let nonarticle_body = DenseTree::generate_textual_format(&document.body_remainder, false);

    let mut tags = seed_lookup
        .look_up(&Url::parse(page_record.url.as_ref())?)?
        .map(|seed: &Seed| seed.tags.iter().map(|cs| cs.to_string()).collect())
        .unwrap_or_else(|| Vec::with_capacity(0));

    // TODO(temporary): add antifeatures as tags, for now.
    let antifeatures = page_record.record.analysed_antifeatures;
    if antifeatures.contains(AnalysisAntifeatures::CLOUDFLARE) {
        tags.push("ZCloudflare".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::ADVERTS) {
        tags.push("ZAdverts".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::ANNOYANCE) {
        tags.push("ZAnnoyance".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::COOKIE_NAG) {
        tags.push("ZCookieNags".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::PRIVACY) {
        tags.push("ZTrackers".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::SOCIAL) {
        tags.push("ZSocialTrackers".to_owned());
    }

    // TODO Store the actual structure of the document in the store?

    let favicon_url_relative = document.head.effective_favicon_url();
    let favicon_url = Url::parse(page_record.url.as_ref())?.join(favicon_url_relative)?;
    let favicon_url = favicon_url.as_str();
    let mut favicon_url_hash_long = [0u8; 8];
    favicon_url_hash_long.copy_from_slice(&blake3::hash(favicon_url.as_bytes()).as_bytes()[0..8]);
    let favicon_url_hash = u64::from_le_bytes(favicon_url_hash_long);

    indexer_backend.add_document(BackendIndependentDocument {
        title: document.head.title,
        article_body,
        nonarticle_body,
        // TODO populate tags & antifeatures
        tags,
        url: page_record.url.to_string(),
        favicon_url_hash,
    })?;

    Ok(())
}

pub fn handle_icon_pack(
    buf_reader: &mut impl BufRead,
    icon_store: &IconStore,
) -> anyhow::Result<()> {
    let page_record: PackRecord<IconEntry> = serde_bare::from_reader(buf_reader)?;

    let webp_bytes = page_record.record.webp_bytes;

    let mut favicon_url_hash_long = [0u8; 8];
    favicon_url_hash_long
        .copy_from_slice(&blake3::hash(page_record.url.as_ref().as_bytes()).as_bytes()[0..8]);
    let favicon_url_hash = u64::from_le_bytes(favicon_url_hash_long);

    icon_store.env.rw_txn(|txn| {
        icon_store
            .icons
            .put(txn, &favicon_url_hash.to_le_bytes(), &webp_bytes)
    })?;

    Ok(())
}

pub struct SeedLookupTable {
    pub by_prefix: PatriciaMap<Seed>,
    pub by_reduced_domain: HashMap<CompactString, Seed>,
}

pub async fn build_seed_lookup_table(
    mut seed_rx: Receiver<Seed>,
) -> anyhow::Result<SeedLookupTable> {
    let mut seed_lookup = SeedLookupTable {
        by_prefix: Default::default(),
        by_reduced_domain: HashMap::new(),
    };

    while let Some(seed) = seed_rx.recv().await {
        match &seed.url {
            UrlOrUrlPattern::Url(url_str) => {
                let url = Url::parse(url_str)?;
                let reduced_domain = get_reduced_domain(&url)?;
                seed_lookup
                    .by_reduced_domain
                    .insert(reduced_domain.into(), seed);
            }
            UrlOrUrlPattern::UrlPrefix(url_str) => {
                seed_lookup.by_prefix.insert(url_str.clone(), seed);
            }
        }
    }

    Ok(seed_lookup)
}

impl SeedLookupTable {
    pub fn look_up(&self, url: &Url) -> anyhow::Result<Option<&Seed>> {
        if let Some((_prefix, seed)) = self.by_prefix.get_longest_common_prefix(&url.as_str()) {
            return Ok(Some(seed));
        }

        let domain = get_reduced_domain(url)?;
        if let Some(seed) = self.by_reduced_domain.get(domain.as_ref()) {
            return Ok(Some(seed));
        }

        Ok(None)
    }
}
