use anyhow::{bail, Context};
use clap::Parser;
use colour::{blue, yellow_ln};
use env_logger::Env;
use itertools::Itertools;
use libmdbx::Database;
use quickpeep_raker::config;
use quickpeep_raker::storage::mdbx_helper_types::{MdbxBare, MdbxString};
use quickpeep_raker::storage::records::OnHoldUrlRecord;
use quickpeep_raker::storage::RakerStore;
use std::collections::HashMap;
use std::path::PathBuf;

/// Seeds a raker's queue with URLs
#[derive(Clone, Debug, Parser)]
pub struct Opts {
    #[clap(long = "config")]
    config: Option<PathBuf>,

    /// Whether to show URLs instead of domains
    #[clap(long = "urls")]
    urls: bool,
}

#[tokio::main]
pub async fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(
        Env::default().default_filter_or("info,quickpeep_raker=debug,qp_raker=debug"),
    )
    .init();

    let opts: Opts = Opts::parse();

    let config_path = opts
        .config
        .unwrap_or_else(|| PathBuf::from("qp_raker.toml"));
    let config = config::RakerConfig::load(&config_path).context("Failed to load config")?;

    if !config.workbench_dir.exists() {
        bail!(
            "Workbench directory ({:?}) doesn't exist.",
            config.workbench_dir
        );
    }

    let store = RakerStore::open(&config.workbench_dir.join("raker.mdbx"))?;
    let is_urls = opts.urls;

    let counts = tokio::task::spawn_blocking(move || -> anyhow::Result<Vec<_>> {
        let counts = count_on_hold(&store, is_urls)?;

        let sorted_counts = counts
            .into_iter()
            .map(|(string, count)| (count, string))
            .sorted()
            .rev()
            .collect_vec();

        Ok(sorted_counts)
    })
    .await??;

    blue!("№ Refs   ");
    if opts.urls {
        yellow_ln!("URL");
    } else {
        yellow_ln!("Domain");
    }
    for (count, string) in counts {
        println!("{:>6}   {}", count, string);
    }

    Ok(())
}

pub fn count_on_hold(store: &RakerStore, urls: bool) -> anyhow::Result<HashMap<String, u32>> {
    let mut map: HashMap<String, u32> = Default::default();

    let txn = store.ro_txn()?;
    let urls_on_hold: &Database = &txn.mdbx.borrow_dbs().urls_on_hold;

    let mut cur = txn.mdbx_txn.cursor(urls_on_hold)?;

    for row in cur.iter_start::<MdbxString, MdbxBare<OnHoldUrlRecord>>() {
        let (domain_then_url, record) = row?;
        let mut split = domain_then_url.0.as_ref().split('\n');
        if urls {
            // Skip one
            split.next();
        }
        let piece = split.next().context("Missing piece")?;
        let count = map.entry(piece.to_owned()).or_insert(0);
        *count += record.0.refs as u32;
    }

    Ok(map)
}
