use chrono::Utc;
use log::warn;
use metrics::{describe_counter, register_counter, Counter, Unit};
use quickpeep_structs::rake_entries::PackRecord;
use reqwest::Url;
use serde::{Deserialize, Serialize};
use std::borrow::Cow;
use std::fs::OpenOptions;
use std::io::Write;
use std::path::Path;
use std::time::Duration;
use tokio::sync::mpsc::Receiver;

/// Size at which a new pack file will be created. 4 GiB, will later be configurable?
pub const SUGGESTED_SIZE_CUTOFF: usize = 4 * 1024 * 1024 * 1024;

/// The Zstd compression level to use. 16 is quite high, but we really want the compact file sizes;
/// willing to pay quite a lot in compression speed.
/// If this turns out to be too slow, should probably go down to 10 or 7.
pub const SUGGESTED_ZSTD_LEVEL: i32 = 16;

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct PackEmitterSettings {
    #[serde(default)]
    pub size_cutoff: Option<usize>,
    #[serde(default)]
    pub zstd_level: Option<i32>,
}

/// An emitter for some kind of pack.
/// Usual types of T: RakedPageEntry, RakedReferrerEntry, PermanentFailure
pub fn pack_emitter<T: Serialize + Send + 'static>(
    directory: &Path,
    name: &str,
    schema_name: &'static str,
    mut rx: Receiver<(Url, T)>,
    settings: &PackEmitterSettings,
) -> anyhow::Result<()> {
    describe_counter!(
        "emitted_pack_bytes",
        Unit::Bytes,
        "Bytes emitted into a pack file (uncompressed)"
    );
    describe_counter!(
        "emitted_pack_count",
        Unit::Count,
        "Records emitted into a pack file"
    );
    loop {
        let now = Utc::now();
        // 2022-01-01 01:01:01
        let new_pack_file_path = loop {
            let new_pack_file_path =
                directory.join(format!("{}.{}.pack", now.format("%F_%T"), name));
            if new_pack_file_path.exists() {
                warn!(
                    "{:?} already exists; sleeping to generate new timestamp.",
                    new_pack_file_path
                );
                std::thread::sleep(Duration::from_secs(2));
            } else {
                break new_pack_file_path;
            }
        };

        if !pack_emitter_to_file(&new_pack_file_path, &mut rx, name, schema_name, settings)? {
            // File wasn't filled; the receiver was exhausted (we're shutting down).
            break;
        }
    }
    Ok(())
}

/// Returns: true if the file was filled (size cutoff reached), false if the receiver was exhausted.
fn pack_emitter_to_file<T: Serialize>(
    file: &Path,
    rx: &mut Receiver<(Url, T)>,
    name: &str,
    schema_name: &'static str,
    settings: &PackEmitterSettings,
) -> anyhow::Result<bool> {
    let file = OpenOptions::new().create_new(true).write(true).open(file)?;
    let mut compressor =
        zstd::stream::Encoder::new(file, settings.zstd_level.unwrap_or(SUGGESTED_ZSTD_LEVEL))?;
    let mut ser_buf = Vec::new();

    let mut length_so_far = 0usize;

    let byte_counter: Counter = register_counter!("emitted_pack_bytes", "pack" => name.to_owned());
    let record_counter: Counter =
        register_counter!("emitted_pack_count", "pack" => name.to_owned());

    serde_bare::to_writer(&mut ser_buf, schema_name)?;

    while let Some((url, record)) = rx.blocking_recv() {
        serde_bare::to_writer(
            &mut ser_buf,
            &PackRecord {
                url: Cow::Borrowed(url.as_str()),
                record,
            },
        )?;

        compressor.write_all(&ser_buf)?;
        length_so_far += ser_buf.len();
        byte_counter.increment(ser_buf.len() as u64);
        record_counter.increment(1);

        if length_so_far > settings.size_cutoff.unwrap_or(SUGGESTED_SIZE_CUTOFF) {
            // MUST CALL
            compressor.finish()?.flush()?;
            return Ok(true);
        }

        ser_buf.clear();
    }

    // MUST CALL
    compressor.finish()?.flush()?;

    // Exhausted
    Ok(false)
}
