use anyhow::{anyhow, Context};
use libmdbx::{Error, TableObject, TransactionKind};
use serde::de::DeserializeOwned;
use serde::Serialize;
use std::borrow::Cow;

/// u32 in native byte endianness (as required by INTEGERKEY mode)
#[derive(Copy, Clone, Debug)]
pub struct MdbxU32(pub u32);

impl MdbxU32 {
    pub fn as_bytes(&self) -> Cow<'_, [u8]> {
        Cow::Owned(self.0.to_ne_bytes().to_vec())
    }
}

impl TableObject<'_> for MdbxU32 {
    fn decode(data_val: &[u8]) -> Result<Self, libmdbx::Error>
    where
        Self: Sized,
    {
        if data_val.len() != 4 {
            return Err(libmdbx::Error::DecodeError(
                anyhow!("MDBX Key not 4 bytes; can't be decoded as u32").into(),
            ));
        }
        let mut buf = [0u8; 4];
        buf.copy_from_slice(&data_val);
        Ok(MdbxU32(u32::from_ne_bytes(buf)))
    }
}

/// u64 in native byte endianness (as required by INTEGERKEY mode)
#[derive(Copy, Clone, Debug)]
pub struct MdbxU64(pub u64);

impl MdbxU64 {
    pub fn as_bytes(&self) -> Cow<'_, [u8]> {
        Cow::Owned(self.0.to_ne_bytes().to_vec())
    }
}

impl TableObject<'_> for MdbxU64 {
    fn decode(data_val: &[u8]) -> Result<Self, libmdbx::Error>
    where
        Self: Sized,
    {
        if data_val.len() != 8 {
            return Err(libmdbx::Error::DecodeError(
                anyhow!("MDBX Key not 8 bytes; can't be decoded as u64").into(),
            ));
        }
        let mut buf = [0u8; 8];
        buf.copy_from_slice(&data_val);
        Ok(MdbxU64(u64::from_ne_bytes(buf)))
    }
}

/// UTF-8 String
#[derive(Clone, Debug)]
pub struct MdbxString<'txn>(pub Cow<'txn, str>);

impl MdbxString<'_> {
    pub fn as_bytes(&self) -> &[u8] {
        self.0.as_bytes()
    }

    pub fn into_string(self) -> String {
        self.0.into_owned()
    }
}

impl<'a> TableObject<'_> for MdbxString<'a> {
    fn decode(_data_val: &[u8]) -> Result<Self, libmdbx::Error>
    where
        Self: Sized,
    {
        unreachable!()
    }

    unsafe fn decode_val<K: TransactionKind>(
        txnptr: *const mdbx_sys::MDBX_txn,
        data_val: &mdbx_sys::MDBX_val,
    ) -> Result<Self, Error>
    where
        Self: Sized,
    {
        let bytes = MdbxBytes::decode_val::<K>(txnptr, data_val)?;
        let string_cow = match bytes {
            Cow::Borrowed(data) => {
                let string = std::str::from_utf8(data)
                    .context("Failed to decode MDBX key as string")
                    .map_err(|e| libmdbx::Error::DecodeError(e.into()))?;
                Cow::Borrowed(string)
            }
            Cow::Owned(data) => {
                let string = String::from_utf8(data)
                    .context("Failed to decode MDBX key as string")
                    .map_err(|e| libmdbx::Error::DecodeError(e.into()))?;
                Cow::Owned(string)
            }
        };
        Ok(MdbxString(string_cow))
    }
}

// /// UTF-8 String
// /// Using Cow<'txn, str> would've needed some unsafe code (see `Cow<'txn, [u8]>` for inspiration),
// /// so I didn't bother for now.
// #[derive(Clone, Debug)]
// pub struct MdbxString(pub String);
//
// impl MdbxString {
//     pub fn as_bytes(&self) -> &[u8] {
//         self.0.as_bytes()
//     }
// }
//
// impl TableObject<'_> for MdbxString {
//     fn decode(data_val: &[u8]) -> Result<Self, libmdbx::Error> where Self: Sized {
//         let string = String::from_utf8(data_val.to_vec())
//             .context("Failed to decode MDBX key as string")
//             .map_err(|e| libmdbx::Error::DecodeError(e.into()))?;
//         Ok(MdbxString(string))
//     }
// }

/// Any BARE payload
#[derive(Clone, Debug)]
pub struct MdbxBare<T>(pub T);

impl<T: Serialize> MdbxBare<T> {
    pub fn as_bytes(&self) -> Vec<u8> {
        serde_bare::to_vec(&self.0).expect("It's unreasonable to expect serialisation will fail")
    }
}

impl<T: DeserializeOwned> TableObject<'_> for MdbxBare<T> {
    fn decode(_data_val: &[u8]) -> Result<Self, libmdbx::Error>
    where
        Self: Sized,
    {
        unreachable!()
    }

    unsafe fn decode_val<K: TransactionKind>(
        txnptr: *const mdbx_sys::MDBX_txn,
        data_val: &mdbx_sys::MDBX_val,
    ) -> Result<Self, Error>
    where
        Self: Sized,
    {
        let bytes = MdbxBytes::decode_val::<K>(txnptr, data_val)?;

        let record = serde_bare::from_slice(bytes.as_ref())
            .context("Failed to decode MDBX key as BARE object")
            .map_err(|e| libmdbx::Error::DecodeError(e.into()))?;

        Ok(MdbxBare(record))
    }
}

/// Supported natively by libmdbx.
pub type MdbxBytes<'txn> = Cow<'txn, [u8]>;
