use crate::raking::{RakeIntent, TemporaryFailure};
use reqwest::Url;
use serde::{Deserialize, Serialize};
use std::collections::BTreeSet;

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct ActiveDomainRecord {
    /// The raffle ticket number owned by this domain.
    pub raffle_ticket: u32,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct UrlVisitedRecord {
    /// Number of minutes since the QuickPeep Epoch that this page was last raked at.
    /// We store minutes to give us 60× the range of times.
    /// We'd really rather stick with 32-bit ints to reduce the space storage requirements.
    /// We could *possibly* go for a u16 in the future and store number of days (179 years' range):
    /// sitemaps and feeds usually only tell you the date the page was last updated.
    pub last_visited_days: u16,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct QueueUrlRecord {
    pub intent: RakeIntent, // TODO CONSIDER
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct OnHoldUrlRecord {
    /// Record that should be emitted once this is released.
    pub queue_record: QueueUrlRecord,

    /// Number of times this URL has been 'enqueued'; capped at 255.
    pub refs: u8,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct BackingOffDomainRecord {
    /// The URL that caused the backoff.
    pub failed_url: String,
    /// The reason that this backoff is in place
    pub failure: TemporaryFailure,
    /// Duration of the backoff. Used to provide increasing backoffs if the failures persist.
    pub backoff: u32,
    /// When the domain should be reinstated
    /// MUST match the timestamp present in the reinstatements table.
    pub reinstate_at: u64,
}

#[derive(Clone, Debug, Serialize, Deserialize, Default)]
pub struct AllowedDomainRecord {
    /// Set of acceptable path prefixes.
    /// Empty if ALL path prefixes are permitted.
    pub restricted_prefixes: BTreeSet<String>,
}

impl AllowedDomainRecord {
    /// Returns true iff this record applies to this URL.
    ///
    /// Preconditions: it has been checked that the record applies to the domain
    pub fn applies_to_url(&self, url: &Url) -> bool {
        if self.restricted_prefixes.is_empty() {
            return true;
        }

        let mut applies = false;
        for prefix in self.restricted_prefixes.iter() {
            if url.path().starts_with(prefix) {
                applies = true;
                break;
            }
            if prefix.as_str() > url.path() {
                // e.g. /dog > /cat/xyz
                // This means we've missed all chances to see our prefix,
                // so we break here (efficiency).
                break;
            }
        }
        applies
    }
}

#[derive(Clone, Debug, Serialize, Deserialize, Default)]
pub struct WeedDomainRecord {
    /// Set of weedy path prefixes.
    /// Empty if ALL path prefixes are weedy.
    pub restricted_prefixes: BTreeSet<String>,
}

impl WeedDomainRecord {
    /// Returns true iff this record applies to this URL.
    ///
    /// Preconditions: it has been checked that the record applies to the domain
    pub fn applies_to_url(&self, url: &Url) -> bool {
        if self.restricted_prefixes.is_empty() {
            return true;
        }

        let mut applies = false;
        for prefix in self.restricted_prefixes.iter() {
            if url.path().starts_with(prefix) {
                applies = true;
                break;
            }
            if prefix.as_str() > url.path() {
                // e.g. /dog > /cat/xyz
                // This means we've missed all chances to see our prefix,
                // so we break here (efficiency).
                break;
            }
        }
        applies
    }
}
